from ..base import DBTestCase
from ..example_project.tables import Band, Manager


class TestInsert(DBTestCase):

    def test_insert(self):
        self.insert_rows()

        Band.insert(
            Band(name='Rustaceans')
        ).run_sync()

        response = Band.select.columns(Band.name).run_sync()
        names = [i['name'] for i in response]

        self.assertTrue(
            'Rustaceans' in names
        )

    def test_add(self):
        self.insert_rows()

        Band.insert().add(
            Band(name='Rustaceans')
        ).run_sync()

        response = Band.select.columns(Band.name).run_sync()
        names = [i['name'] for i in response]

        self.assertTrue(
            'Rustaceans' in names
        )

    def test_incompatible_type(self):
        """
        You shouldn't be able to add instances of a different table.
        """
        with self.assertRaises(TypeError):
            Band.insert().add(
                Manager(name="Guido")
            ).run_sync()



from __future__ import annotations

from piccolo.query.base import Query
from piccolo.query.mixins import LimitMixin, OrderByMixin, WhereMixin, Output
from piccolo.querystring import QueryString
from .select import Select


class Objects(
    Query,
    LimitMixin,
    OrderByMixin,
    WhereMixin,
):
    """
    Almost identical to select, except you have to select all fields, and
    table instances are returned, rather than just data.
    """

    _output = Output(as_objects=True)

    @property
    def querystring(self) -> QueryString:
        select = Select(
            table=self.table,
            column_names=[]
        )

        for attr in ('_limit', '_where', '_output', 'order_by'):
            setattr(select, attr, getattr(self, attr))

        return select.querystring

    def __str__(self) -> str:
        return self.querystring.__str__()




import asyncio
import typing as t

import asyncpg
from asyncpg.pool import Pool

from piccolo.engine.base import Engine
from piccolo.query.base import Query
from piccolo.querystring import QueryString
from piccolo.utils.sync import run_sync


class Transaction():
    """
    Usage:

    transaction = engine.Transaction()
    transaction.add(Foo.create)
    transaction.run_sync()
    """

    def __init__(self, engine):
        self.engine = engine
        self.queries = []

    def add(self, *query: Query):
        self.queries += list(query)

    async def _run_queries(self, connection):
        async with connection.transaction():
            for query in self.queries:
                await connection.execute(query.__str__())

        self.queries = []

    async def _run_in_pool(self):
        pool = await self.engine.get_pool()
        connection = await pool.acquire()

        try:
            await self._run_queries(connection)
        except Exception:
            pass
        finally:
            await pool.release(connection)

    async def _run(self):
        connection = await asyncpg.connect(**self.engine.config)
        await self._run_queries(connection)

    async def run(self):
        await self._run_in_pool()

    def run_sync(self):
        return run_sync(
            self._run()
        )


class PostgresEngine(Engine):
    """
    Currently when using run ...  it sets up a connection each time.

    When instantiated ... create the connection pool ...

    Needs to be a singleton that's shared by all the tables.
    """

    engine_type = 'postgres'

    def __init__(self, config: t.Dict[str, t.Any]) -> None:
        self.config = config
        self.pool: t.Optional[Pool] = None
        self.loop: t.Optional[asyncio.AbstractEventLoop] = None

    async def get_pool(self) -> Pool:
        loop = asyncio.get_event_loop()
        if not self.pool or (self.loop != loop):
            self.pool = await asyncpg.create_pool(
                **self.config
            )
            self.loop = loop
        return self.pool

    async def run_in_pool(self, query: str, args: t.List[t.Any] = []):
        pool = await self.get_pool()

        connection = await pool.acquire()
        try:
            response = await connection.fetch(query, *args)
        except Exception:
            pass
        finally:
            await pool.release(connection)

        return response

    async def run(self, query: str, args: t.List[t.Any] = []):
        connection = await asyncpg.connect(**self.config)
        results = await connection.fetch(query, *args)
        await connection.close()
        return results

    async def run_querystring(
        self,
        querystring: QueryString,
        in_pool: bool = False
    ):
        if in_pool:
            return await self.run_in_pool(
                *querystring.compile_string(engine_type=self.engine_type)
            )
        else:
            return await self.run(
                *querystring.compile_string(engine_type=self.engine_type)
            )

    def transaction(self):
        return Transaction(engine=self)



from __future__ import annotations

from piccolo.query.base import Query
from piccolo.querystring import QueryString


class Create(Query):
    """
    Creates a database table.
    """

    @property
    def querystring(self) -> QueryString:
        base = f'CREATE TABLE {self.table.Meta.tablename}'
        columns = ', '.join(['{}' for i in self.table.Meta.columns])
        query = f'{base} ({columns})'
        return QueryString(
            query,
            *[i.querystring for i in self.table.Meta.columns]
        )

    def __str__(self) -> str:
        return self.querystring.__str__()



from __future__ import annotations
import dataclasses
import itertools

from piccolo.columns.base import Column
from piccolo.query.base import Query
from piccolo.querystring import QueryString


@dataclasses.dataclass
class Rename():
    """
    Band.alter.rename(Band.popularity, ‘rating’)
    """
    column: Column
    new_name: str

    @property
    def querystring(self) -> QueryString:
        return QueryString(
            f'RENAME {self.column._name} TO {self.new_name}',
        )

    def __str__(self) -> str:
        return self.querystring.__str__()


@dataclasses.dataclass
class Drop():
    """
    Band.alter.drop('popularity')
    """
    column: Column

    @property
    def querystring(self) -> QueryString:
        return QueryString(
            f'DROP {self.column._name}'
        )

    def __str__(self) -> str:
        return self.querystring.__str__()


@dataclasses.dataclass
class Add():
    """
    Band.alter.add(‘members’, Integer())
    """
    name: str
    column: Column

    @property
    def querystring(self) -> QueryString:
        self.column._name = self.name
        return QueryString(
            'ADD {}',
            self.column.querystring
        )

    def __str__(self) -> str:
        return self.querystring.__str__()


class Alter(Query):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._add: t.List[Add] = []
        self._drop: t.List[Drop] = []
        self._rename: t.List[Rename] = []

    def add(self, name: str, column: Column) -> Alter:
        self._add.append(
            Add(name, column)
        )
        return self

    def rename(self, column: Column, new_name: str) -> Alter:
        self._rename.append(
            Rename(column, new_name)
        )
        return self

    def drop(self, column: Column) -> Alter:
        self._drop.append(
            Drop(column)
        )
        return self

    @property
    def querystring(self) -> QueryString:
        query = f'ALTER TABLE {self.table.Meta.tablename}'

        alterations = [
            i.querystring for i in itertools.chain(
                self._add,
                self._rename,
                self._drop
            )
        ]

        for a in alterations:
            query += ' {}'

        return QueryString(
            query,
            *alterations
        )

    def __str__(self) -> str:
        return self.querystring.__str__()



import datetime
import typing as t


if t.TYPE_CHECKING:
    from columns.combination import Where, And, Or  # noqa


Combinable = t.Union['Where', 'And', 'Or']


Iterable = t.Iterable[t.Any]


Datetime = t.Union[
    datetime.datetime,
    t.Callable[[], datetime.datetime],
    None,
]


__all__ = (
    'Combinable',
    'Iterable',
    'Datetime'
)



from unittest import TestCase

from ..example_project.tables import Band, Manager


class TestTransaction(TestCase):

    def test_error(self):
        """
        Make sure queries in a transaction aren't committed if a query fails.
        """
        transaction = Band.Meta.db.transaction()
        transaction.add(
            Manager.create,
            Band.create,
            Band.raw('MALFORMED QUERY ... SHOULD ERROR')
        )
        try:
            transaction.run_sync()
        except Exception:
            pass
        self.assertTrue(
            not Band.table_exists.run_sync()
        )
        self.assertTrue(
            not Manager.table_exists.run_sync()
        )

    def test_succeeds(self):
        transaction = Band.Meta.db.transaction()
        transaction.add(
            Manager.create,
            Band.create
        )
        transaction.run_sync()

        self.assertTrue(
            Band.table_exists.run_sync()
        )
        self.assertTrue(
            Manager.table_exists.run_sync()
        )

        transaction.add(
            Band.drop,
            Manager.drop
        )
        transaction.run_sync()



from __future__ import annotations

from piccolo.query.base import Query
from piccolo.query.mixins import AddMixin
from piccolo.querystring import QueryString


class Insert(Query, AddMixin):

    def run_callback(self, results):
        """
        Assign the ids of the created rows to the model instances.
        """
        for index, row in enumerate(results):
            self._add[index].id = row['id']

    @property
    def sqlite_querystring(self) -> QueryString:
        if len(self._add) > 1:
            raise Exception(
                "The SQLite engine is unable to insert more than one row at a "
                "time."
            )

        base = f'INSERT INTO {self.table.Meta.tablename}'
        columns = ','.join(
            [i._name for i in self.table.Meta.columns]
        )
        values = ','.join([
            '{}' for i in self._add
        ])
        query = f'{base} ({columns}) VALUES {values}'
        return QueryString(
            query,
            *[i.querystring for i in self._add],
            query_type='insert'
        )

    @property
    def postgres_querystring(self) -> QueryString:
        base = f'INSERT INTO {self.table.Meta.tablename}'
        columns = ','.join(
            [i._name for i in self.table.Meta.columns]
        )
        values = ','.join([
            '{}' for i in self._add
        ])
        query = f'{base} ({columns}) VALUES {values} RETURNING id'
        return QueryString(
            query,
            *[i.querystring for i in self._add],
            query_type='insert'
        )



import json

from ..base import DBTestCase
from ..example_project.tables import Band


class TestOutput(DBTestCase):

    def test_output_as_list(self):
        self.insert_row()

        response = Band.select.columns(
            Band.name
        ).output(
            as_list=True
        ).run_sync()

        self.assertTrue(
            response == ['Pythonistas']
        )

    def test_output_as_json(self):
        self.insert_row()

        response = Band.select.columns(
            Band.name
        ).output(
            as_json=True
        ).run_sync()

        self.assertTrue(
            json.loads(response) == [{'name': 'Pythonistas'}]
        )



from unittest import TestCase

from tests.example_project.tables import Manager


class TestDelete(TestCase):

    def setUp(self):
        Manager.create.run_sync()

    def tearDown(self):
        Manager.drop.run_sync()

    def test_delete(self):
        manager = Manager(
            name='Maz'
        )

        manager.save.run_sync()
        manager.remove.run_sync()

        # how can I implement 'flat=True'
        # Band.select.columns(Band.name).output(as_list=True).run_sync()
        #
        Manager.select.columns(Manager.name).run_sync()



from ..base import DBTestCase
from ..example_project.tables import Band


class TestInstance(DBTestCase):
    """
    Test instantiating Table instances
    """

    def test_insert(self):
        Pythonistas = Band(name="Pythonistas")
        self.assertEqual(Pythonistas.__str__(), "('Pythonistas',null,null,DEFAULT)")

    def test_non_existant_column(self):
        with self.assertRaises(ValueError):
            Band(name="Pythonistas", foo="bar")



from piccolo import table
from piccolo import columns
from piccolo.engine import PostgresEngine


DB = PostgresEngine({
    'host': 'localhost',
    'database': 'piccolo',
    'user': 'piccolo',
    'password': 'piccolo'
})


###############################################################################
# Simple example

class Manager(table.Table):
    name = columns.Varchar(length=50)

    class Meta():
        db = DB


class Band(table.Table):
    name = columns.Varchar(length=50)
    manager = columns.ForeignKey(Manager)
    popularity = columns.Integer()

    class Meta():
        db = DB


###############################################################################
# More complex

class Venue(table.Table):
    name = columns.Varchar(length=100)
    capacity = columns.Integer()

    class Meta():
        db = DB


class Concert(table.Table):
    band_1 = columns.ForeignKey(Band)
    band_2 = columns.ForeignKey(Band)
    venue = columns.ForeignKey(Venue)

    class Meta():
        db = DB



import typing as t

from .operators import (
    Equal,
    GreaterEqualThan,
    GreaterThan,
    In,
    LessEqualThan,
    LessThan,
    Like,
    NotEqual,
    NotIn,
    NotLike
)
from .combination import Where
from ..custom_types import Iterable
from ..querystring import QueryString

if t.TYPE_CHECKING:
    from ..table import Table  # noqa
    from .column_types import ForeignKey  # noqa


class Column():

    def __init__(
        self,
        null: bool = True,
        primary: bool = False,
        key: bool = False
    ) -> None:
        self.null = null
        self.primary = primary
        self.key = key
        # Set by Table metaclass:
        self._name: t.Optional[str] = None
        # Used by foreign keys:
        self.call_chain: t.List['ForeignKey'] = []
        self.table_alias: t.Optional[str] = None

    def is_in(self, values: Iterable) -> Where:
        return Where(column=self, values=values, operator=In)

    def not_in(self, values: Iterable) -> Where:
        return Where(column=self, values=values, operator=NotIn)

    def like(self, value: str) -> Where:
        if '%' not in value:
            raise ValueError('% is required for like operators')
        return Where(column=self, value=value, operator=Like)

    def not_like(self, value: str) -> Where:
        if '%' not in value:
            raise ValueError('% is required for like operators')
        return Where(column=self, value=value, operator=NotLike)

    def __lt__(self, value) -> Where:
        return Where(column=self, value=value, operator=LessThan)

    def __le__(self, value) -> Where:
        return Where(column=self, value=value, operator=LessEqualThan)

    def __gt__(self, value) -> Where:
        return Where(column=self, value=value, operator=GreaterThan)

    def __ge__(self, value) -> Where:
        return Where(column=self, value=value, operator=GreaterEqualThan)

    def __eq__(self, value) -> Where:  # type: ignore
        return Where(column=self, value=value, operator=Equal)

    def __ne__(self, value) -> Where:  # type: ignore
        return Where(column=self, value=value, operator=NotEqual)

    def __hash__(self):
        return hash(self._name)

    def get_full_name(self, just_alias=False) -> str:
        """
        Returns the full column name, taking into account joins.
        """
        column_name = self._name

        if not self.call_chain:
            return f"{self._table.Meta.tablename}.{column_name}"

        column_name = (
            "$".join([i._name for i in self.call_chain])
            + f"${column_name}"
        )
        alias = f"{self.call_chain[-1].table_alias}.{self._name}"
        if just_alias:
            return alias
        else:
            return f'{alias} AS "{column_name}"'

    @property
    def querystring(self) -> QueryString:
        """
        Used when creating tables.
        """
        name = getattr(self, '_name', '')
        column_type = getattr(
            self,
            'column_type',
            self.__class__.__name__.upper()
        )
        query = f'{name} {column_type}'
        if self.primary:
            query += ' PRIMARY'
        if self.key:
            query += ' KEY'

        references = getattr(self, 'references', None)
        if references:
            query += f' REFERENCES {references.Meta.tablename}'

        return QueryString(query)

    def __str__(self):
        return self.querystring.__str__()



from __future__ import annotations
import dataclasses
import typing as t

from piccolo.columns import And, Column, Where, Or
from piccolo.custom_types import Combinable
from piccolo.querystring import QueryString

if t.TYPE_CHECKING:
    from table import Table  # noqa


class Limit():

    def __init__(self, number: int) -> None:
        if type(number) != int:
            raise TypeError('Limit must be an integer')
        self.number = number

    @property
    def querystring(self) -> QueryString:
        return QueryString(
            f' LIMIT {self.number}'
        )

    def __str__(self) -> str:
        return self.querystring.__str__()


@dataclasses.dataclass
class OrderBy():
    columns: t.Tuple[Column]
    ascending: bool

    @property
    def querystring(self) -> QueryString:
        order = 'ASC' if self.ascending else 'DESC'
        columns_names = ', '.join(
            [i.get_full_name(just_alias=True) for i in self.columns]
        )
        return QueryString(
            f' ORDER BY {columns_names} {order}'
        )

    def __str__(self):
        return self.querystring.__str__()


@dataclasses.dataclass
class Output():
    as_json: bool = False
    as_list: bool = False
    as_objects: bool = False


class WhereMixin():

    def __init__(self):
        super().__init__()
        self._where: t.Optional[Combinable] = []
        self._where_columns: t.List[Column] = []

    def get_where_columns(self):
        """
        Retrieves all columns used in the where clause - in case joins are
        needed.
        """
        self._where_columns = []
        self._extract_columns(self._where)
        return self._where_columns

    def _extract_columns(self, combinable: Combinable):
        if isinstance(combinable, Where):
            self._where_columns.append(combinable.column)
        elif isinstance(combinable, And) or isinstance(combinable, Or):
            self._extract_columns(combinable.first)
            self._extract_columns(combinable.second)

    def where(self, where: Combinable):
        if self._where:
            self._where = And(self._where, where)
        else:
            self._where = where
        return self


class OrderByMixin():

    def __init__(self):
        super().__init__()
        self._order_by: t.Optional[OrderBy] = None

    def get_order_by_columns(self) -> t.List[Column]:
        return list(self._order_by.columns) if self._order_by else []

    def order_by(self, *columns: Column, ascending=True):
        self._order_by = OrderBy(columns, ascending)
        return self


class LimitMixin():

    def __init__(self):
        super().__init__()
        self._limit: t.Optional[Limit] = None

    def limit(self, number: int):
        self._limit = Limit(number)
        return self

    @staticmethod
    def _response_handler(response):
        if len(response) == 0:
            raise ValueError('No results found')

        return response[0]

    @property
    def first(self):
        self._limit = Limit(1)
        self.response_handler = self._response_handler
        return self


class DistinctMixin():

    def __init__(self):
        super().__init__()
        self._distinct: bool = False

    def distinct(self):
        self._distinct = True
        return self


class CountMixin():

    def __init__(self):
        super().__init__()
        self._count: bool = False

    def count(self):
        self._count = True
        return self


class AddMixin():

    def __init__(self):
        super().__init__()
        self._add: t.List[Table] = []

    def add(self, *instances: Table):
        for instance in instances:
            if not isinstance(instance, self.table):
                raise TypeError('Incompatible type added.')
        self._add += instances
        return self


class OutputMixin():
    """
    Example usage:

    .output(as_list=True)
    .output(as_json=True)
    .output(as_json=True, as_list=True)
    """

    def __init__(self):
        super().__init__()
        self._output = Output()

    def output(
        self,
        as_list: t.Optional[bool] = None,
        as_json: t.Optional[bool] = None
    ):
        if type(as_list) is bool:
            self._output.as_list = as_list

        if type(as_json) is bool:
            self._output.as_json = as_json

        return self


class ColumnsMixin():
    """
    Example usage:

    .columns(MyTable.column_a, MyTable.column_b)
    """

    def __init__(self):
        super().__init__()
        self.selected_columns: t.List[Column] = []

    def columns(self, *columns: Column):
        self.selected_columns += columns
        return self


class ValuesMixin():
    """
    Used to specify new column values - primarily used in update queries.

    Example usage:

    .values({MyTable.column_a: 1})
    """

    def __init__(self):
        super().__init__()
        self._values: t.Dict[Column, t.Any] = {}

    def values(self, values: t.Dict[Column, t.Any]):
        self._values.update(values)
        return self



from .alter import Alter
from .create import Create
from .delete import Delete
from .drop import Drop
from .exists import Exists
from .insert import Insert
from .objects import Objects
from .raw import Raw
from .select import Select
from .table_exists import TableExists
from .update import Update


__all__ = (
    'Alter',
    'Create',
    'Delete',
    'Drop',
    'Exists',
    'Insert',
    'Objects',
    'Raw',
    'Select',
    'TableExists',
    'Update'
)



from unittest import TestCase
from ..example_project.tables import Manager


class TestCreate(TestCase):

    def test_create_table(self):
        Manager.create.run_sync()

        # Just do a count to make sure the table was created ok.
        count_query = Manager.select.count()
        response = count_query.run_sync()

        self.assertEqual(response[0]['count'], 0)

        Manager.drop.run_sync()



from __future__ import annotations
import copy
import typing as t

from piccolo.engine import Engine
from piccolo.columns import Column, PrimaryKey, ForeignKey
from piccolo.query import (
    Alter,
    Create,
    Delete,
    Drop,
    Exists,
    Insert,
    Objects,
    Raw,
    Select,
    TableExists,
    Update,
)
from piccolo.querystring import QueryString, Unquoted
from piccolo.utils import _camel_to_snake


class TableMeta(type):

    def __new__(cls, name, bases, namespace, **kwds):
        """
        Automatically populate the tablename, and columns.
        """
        table = super().__new__(cls, name, bases, namespace)

        Meta = namespace.get('Meta')
        if (not Meta) or (Meta in [i.Meta for i in bases]):
            # We're inheritting a Meta from a parent class - give this class
            # its own Meta
            Meta = type('Meta', tuple(), {})
            table.Meta = Meta

        tablename = getattr(Meta, 'tablename', None) if Meta else None
        if not tablename:
            table.Meta.tablename = _camel_to_snake(name)

        columns = []

        # In case super classes also have columns.
        if bases:
            for base in bases:
                if hasattr(base, 'Meta'):
                    _columns = getattr(base.Meta, 'columns', None)
                    if _columns:
                        # Don't copy over the primary key.
                        columns = [i for i in _columns if not
                                   isinstance(i, PrimaryKey)] + columns

        primary_key = PrimaryKey()
        namespace['id'] = primary_key
        table.id = primary_key

        for key, value in namespace.items():
            if issubclass(type(value), Column):
                columns.append(value)
                value._name = key
                value._table = table

        table.Meta.columns = columns
        table.Meta.non_default_columns = [
            i for i in columns if type(i) != PrimaryKey
        ]
        return table

    ###########################################################################

    def __str__(cls):
        """
        Returns a basic string representation of the table and its columns.

        Used by the playground.
        """
        spacer = '\n    '
        columns = []
        for col in cls.Meta.columns:
            if type(col) == ForeignKey:
                columns.append(
                    f'{col._name} = ForeignKey({col.references.__name__})'
                )
            else:
                columns.append(f'{col._name} = {col.__class__.__name__}()')
        columns_string = spacer.join(columns)
        return (
            f'class {cls.__name__}(Table):\n'
            f'    {columns_string}\n'
        )

    ###########################################################################
    # Class properties
    # TODO - might need to rework, tab completion only works in some versions
    # of iPython.

    @property
    def delete(cls) -> Delete:
        """
        await Band.delete.where(Band.name == 'CSharps').run()
        """
        return Delete(
            table=cls
        )

    @property
    def create(cls) -> Create:
        """
        Create table, along with all columns.

        await Band.create.run()
        """
        return Create(
            table=cls,
        )

    @property
    def create_without_columns(cls) -> Raw:
        """
        Create the table, but with no columns (useful for migrations).

        await Band.create.run()
        """
        return Raw(
            table=cls,
            base=f'CREATE TABLE "{cls.Meta.tablename}"()'
        )

    @property
    def drop(cls) -> Drop:
        """
        Drops the table.

        await Band.drop.run()
        """
        return Drop(
            table=cls,
        )

    @property
    def alter(cls) -> Alter:
        """
        await Band.alter.rename(Band.popularity, 'rating')
        """
        return Alter(
            table=cls,
        )

    @property
    def objects(cls) -> Objects:
        return Objects(
            table=cls
        )

    @property
    def exists(cls) -> Exists:
        """
        Use it to check if a row exists ... not if the table exists.
        """
        return Exists(
            table=cls,
        )

    @property
    def table_exists(cls) -> TableExists:
        return TableExists(
            table=cls,
        )

    @property
    def select(cls) -> Select:
        """
        Get data.

        await Band.select.columns(Band.name).run()
        """
        return Select(
            table=cls,
        )

    @property
    def update(cls) -> Update:
        """
        Update rows.

        await Band.update.values(
            {Band.name: "Spamalot"}
        ).where(Band.name=="Pythonistas")
        """
        return Update(
            table=cls,
        )


class Table(metaclass=TableMeta):

    class Meta:
        tablename = None
        columns: t.List[Column] = []
        non_default_columns: t.List[Column] = []
        db: t.Optional[Engine] = None

    def __init__(self, **kwargs):
        """
        TODO:
        Need to know the memory size of each instance ... can't be
        massive.
        """
        for column in self.Meta.columns:
            value = kwargs.pop(column._name, None)
            if not value:
                if column.default:
                    # Can't use inspect - can't tell that datetime.datetime.now
                    # is a callable.
                    is_callable = hasattr(column.default, '__call__')
                    value = column.default(
                        engine_type=self.Meta.db.engine_type
                    ) if is_callable else column.default
                else:
                    if not column.null:
                        raise ValueError(f"{column._name} wasn't provided")
            self[column._name] = value

        unrecognized = kwargs.keys()
        if unrecognized:
            raise ValueError(f'Unrecognized columns - {unrecognized}')

    @property
    def save(self):
        """
        A proxy to an insert or update query.
        """
        if not hasattr(self, 'id'):
            raise ValueError('No id value found')

        cls = self.__class__

        if type(self.id) == int:
            # pre-existing row
            kwargs = {
                i: getattr(self, i._name, None) for i in cls.Meta.columns
            }
            _id = kwargs.pop('id')
            return cls.update.values(kwargs).where(
                cls.id == _id
            )
        else:
            return cls.insert().add(self)

    @property
    def remove(self):
        """
        A proxy to a delete query.
        """
        _id = self.id

        if type(_id) != int:
            raise ValueError('Can only delete pre-existing rows with an id.')

        self.id = None

        return self.__class__.delete.where(
            self.__class__.id == _id
        )

    def get_related(self, column_name: str):
        """
        some_band.get_related('manager')
        """
        cls = self.__class__

        foreign_key = cls.get_column_by_name(column_name)  # type: ignore
        references = foreign_key.references

        return references.objects.where(
            references.get_column_by_name('id') == getattr(self, column_name)
        ).first

    def __setitem__(self, key: str, value: t.Any):
        setattr(self, key, value)

    def __getitem__(self, key: str):
        return getattr(self, key)

    ###########################################################################

    @property
    def querystring(self) -> QueryString:
        """
        Used when inserting rows.
        """
        args_dict = {
            col._name: self[col._name] for col in self.Meta.columns
        }

        is_unquoted = (lambda arg: type(arg) == Unquoted)

        # Strip out any args which are unquoted.
        # TODO Not the cleanest place to have it (would rather have it handled
        # in the Querystring bundle logic) - might need refactoring.
        filtered_args = [
            i for i in args_dict.values() if not is_unquoted(i)
        ]

        # If unquoted, dump it straight into the query.
        query = ",".join([
            args_dict[column._name].value if is_unquoted(
                args_dict[column._name]
            ) else '{}' for column in self.Meta.columns
        ])
        return QueryString(
            f'({query})',
            *filtered_args
        )

    def __str__(self) -> str:
        return self.querystring.__str__()

    ###########################################################################

    @classmethod
    def get_column_by_name(cls, column_name: str) -> Column:
        columns = [i for i in cls.Meta.columns if i._name == column_name]

        if len(columns) != 1:
            raise ValueError(
                f"Can't find a column called {column_name}."
            )

        return columns[0]

    @classmethod
    def ref(
        cls,
        column_name: str
    ) -> Column:
        """
        Used to get a copy of a column in a reference table.

        Example: manager.name
        """
        local_column_name, reference_column_name = column_name.split('.')

        local_column = cls.get_column_by_name(local_column_name)

        if not isinstance(local_column, ForeignKey):
            raise ValueError(f"{local_column_name} isn't a ForeignKey")

        reference_column = local_column.references.get_column_by_name(
            reference_column_name
        )

        _reference_column = copy.deepcopy(reference_column)
        _reference_column.name = f'{local_column_name}.{reference_column_name}'
        return _reference_column

    ###########################################################################

    @classmethod
    # TODO - needs refactoring into Band.insert.rows(some_table_instance)
    def insert(cls, *rows: 'Table') -> Insert:
        """
        await Band.insert(
            Band(name="Pythonistas", popularity=500, manager=1)
        ).run()
        """
        query = Insert(
            table=cls,
        )
        if rows:
            query.add(*rows)
        return query

    @classmethod
    def raw(cls, sql: str) -> Raw:
        """
        await Band.raw('select * from foo')
        """
        return Raw(
            table=cls,
            base=QueryString(sql)
        )



from unittest import TestCase

from tests.example_project.tables import Manager


class TestSave(TestCase):

    def setUp(self):
        Manager.create.run_sync()

    def tearDown(self):
        Manager.drop.run_sync()

    def test_save_new(self):
        """
        Make sure that saving a new instance works.
        """
        manager = Manager(
            name='Maz',
        )

        query = manager.save
        print(query)
        self.assertTrue('INSERT' in query.__str__())

        query.run_sync()

        names = [
            i['name'] for i in Manager.select.columns(Manager.name).run_sync()
        ]
        self.assertTrue('Maz' in names)

        manager.name = 'Maz2'
        query = manager.save
        print(query)
        self.assertTrue('UPDATE' in query.__str__())

        query.run_sync()
        names = [
            i['name'] for i in Manager.select.columns(Manager.name).run_sync()
        ]
        self.assertTrue('Maz2' in names)
        self.assertTrue('Maz' not in names)

        # Make sure it has an id too ...
        # and the next query is an UPDATE ...
        # .run(objects=True)
        # .run(as_json=True)
        # Generic('some_table').select('name')
        # -> won't work because where requires things like
        # Band.select().where(Band.id == 1)
        # So doesn't work as a generic query build without first defining
        # the tables.... this is totally fine though.



from __future__ import annotations

from piccolo.query.base import Query
from piccolo.query.mixins import WhereMixin
from piccolo.querystring import QueryString


class Delete(Query, WhereMixin):

    @property
    def querystring(self) -> QueryString:
        query = f'DELETE FROM {self.table.Meta.tablename}'
        if self._where:
            query += ' WHERE {}'
            return QueryString(
                query,
                self._where.querystring
            )
        else:
            return QueryString(query)

    def __str__(self) -> str:
        query = f'DELETE FROM {self.table.Meta.tablename}'
        if self._where:
            query += f' WHERE {self._where.__str__()}'
        return query



from __future__ import annotations

from piccolo.query.base import Query
from piccolo.query.mixins import WhereMixin
from piccolo.querystring import QueryString
from .select import Select


class Exists(Query, WhereMixin):

    @property
    def querystring(self) -> QueryString:
        select = Select(
            self.table,
            QueryString(f'SELECT * FROM {self.table.Meta.tablename}')
        )
        select._where = self._where
        return QueryString('SELECT EXISTS({})', select.querystring)

    def __str__(self) -> str:
        return self.querystring.__str__()



from __future__ import annotations

from piccolo.query.base import Query
from piccolo.query.mixins import ValuesMixin, WhereMixin
from piccolo.querystring import QueryString


class Update(Query, ValuesMixin, WhereMixin):

    @property
    def sqlite_querystring(self) -> QueryString:
        raise NotImplementedError()

    def validate(self):
        if len(self._values) == 0:
            raise ValueError(
                "No values were specified to update - please use .values"
            )

        for column, value in self._values.items():
            if len(column.call_chain) > 0:
                raise ValueError(
                    "Related values can't be updated via an update"
                )

    @property
    def postgres_querystring(self) -> QueryString:
        self.validate()

        columns_str = ', '.join([
            f'{col._name} = {{}}' for col, _ in self._values.items()
        ])

        query = f'UPDATE {self.table.Meta.tablename} SET ' + columns_str

        querystring = QueryString(
            query,
            *self._values.values()
        )

        # TODO - need to support joins
        # UPDATE band SET name='bob' FROM manager WHERE band.manager = manager.id AND manager.name = 'Guido';
        # I think the joins should work differently from select i.e. no need
        # to shared underlying code.

        if self._where:
            where_querystring = QueryString(
                '{} WHERE {}',
                querystring,
                self._where.querystring
            )
            return where_querystring
        else:
            return querystring




from __future__ import annotations

from piccolo.query.base import Query
from piccolo.querystring import QueryString


class TableExists(Query):

    def response_handler(self, response):
        return response[0]['exists']

    @property
    def querystring(self) -> QueryString:
        return QueryString(
            "SELECT EXISTS(SELECT * FROM information_schema.tables WHERE "
            f"table_name = '{self.table.Meta.tablename}')"
        )

    def __str__(self) -> str:
        return self.querystring.__str__()



import copy
import typing as t

from piccolo.columns.base import Column
from piccolo.querystring import Unquoted

if t.TYPE_CHECKING:
    import table  # noqa
    from piccolo.custom_types import Datetime  # noqa


class Varchar(Column):

    def __init__(self, length: int = 255, default: str = None,
                 **kwargs) -> None:
        self.length = length
        self.default = default
        super().__init__(**kwargs)


class Integer(Column):

    def __init__(self, default: int = None, **kwargs) -> None:
        self.default = default
        super().__init__(**kwargs)


class Serial(Column):

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)


DEFAULT = Unquoted('DEFAULT')
NULL = Unquoted('null')


class PrimaryKey(Column):

    @property
    def column_type(self):
        engine_type = self._table.Meta.db.engine_type
        if engine_type == 'postgres':
            return 'SERIAL'
        elif engine_type == 'sqlite':
            return 'INTEGER'
        raise Exception('Unrecognized engine type')

    def default(self, engine_type: str = 'postgres'):
        if engine_type == 'postgres':
            return DEFAULT
        elif engine_type == 'sqlite':
            return NULL
        else:
            raise Exception('Unrecognized engine type')

    def __init__(self, **kwargs) -> None:
        kwargs.update({
            'primary': True,
            'key': True
        })
        super().__init__(**kwargs)


class Timestamp(Column):

    def __init__(self, default: 'Datetime' = None, **kwargs) -> None:
        self.default = default
        super().__init__(**kwargs)


class Boolean(Column):

    def __init__(self, default: bool = False, **kwargs) -> None:
        self.default = default
        super().__init__(**kwargs)


class ForeignKey(Integer):
    """
    Returns an integer, representing the referenced row's ID.

        some_band.manager
        >>> 1

    Can also use it to perform joins:

        Band.select.columns(Band.manager.name)

    To get a referenced row as an object:

        User.object.where(User.id == some_band.manager)

    Or use this, which is just a proxy to the above:

        some_band.related_object('manager')

    To change the manager:

        some_band.manager = some_manager_id
        some_band.save

    Or:
        some_band.set_related_object('manager', some_manager)

    """

    column_type = 'INTEGER'

    def __getattribute__(self, name):
        """
        Returns attributes unmodified unless they're Column instances, in which
        case a copy is returned with an updated call_chain (which records the
        joins required).
        """
        # see if it has an attribute with that name ...
        # if it's asking for a foreign key ... return a copy of self ...

        try:
            value = object.__getattribute__(self, name)
        except AttributeError:
            raise AttributeError

        foreignkey_class = object.__getattribute__(self, '__class__')

        if type(value) == foreignkey_class:  # i.e. a ForeignKey
            new_column = copy.deepcopy(value)
            new_column.call_chain = copy.copy(self.call_chain)
            new_column.call_chain.append(self)

            # We have to set limits to the call chain because Table 1 can
            # reference Table 2, which references Table 1, creating an endless
            # loop. For now an arbitrary limit is set of 10 levels deep.
            # When querying a call chain more than 10 levels deep, an error
            # will be raised. Often there are more effective ways of
            # structuring a query than joining so many tables anyway.
            if len(new_column.call_chain) > 10:
                raise Exception('Call chain too long!')

            for proxy_column in self.proxy_columns:
                try:
                    delattr(new_column, proxy_column._name)
                except Exception:
                    pass

            for column in value.references.Meta.columns:
                _column = copy.deepcopy(column)
                _column.call_chain = copy.copy(new_column.call_chain)
                _column.call_chain.append(new_column)
                if _column._name == 'id':
                    continue
                setattr(new_column, _column._name, _column)
                self.proxy_columns.append(_column)

            return new_column
        elif issubclass(type(value), Column):
            new_column = copy.deepcopy(value)
            new_column.call_chain = copy.copy(self.call_chain)
            new_column.call_chain.append(self)
            return new_column
        else:
            return value

    def __init__(self, references: t.Type['table.Table'], **kwargs) -> None:
        super().__init__(**kwargs)
        self.references = references
        self.proxy_columns: t.List[Column] = []

        # Allow columns on the referenced table to be accessed via auto
        # completion.
        for column in references.Meta.columns:
            _column = copy.deepcopy(column)
            setattr(self, column._name, _column)
            self.proxy_columns.append(_column)



from ..base import DBTestCase
from ..example_project.tables import Band


class TestUpdate(DBTestCase):

    def test_update(self):
        self.insert_rows()

        Band.update.values({
            Band.name: 'Pythonistas3'
        }).where(
            Band.name == 'Pythonistas'
        ).run_sync()

        response = Band.select.columns(
            Band.name
        ).where(
            Band.name == 'Pythonistas3'
        ).run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'Pythonistas3'}]
        )



from __future__ import annotations
import typing as t

from .operators import Operator
from ..custom_types import Combinable, Iterable
from piccolo.querystring import QueryString

if t.TYPE_CHECKING:
    from .base import Column  # noqa


class CombinableMixin(object):
    def __and__(self, value: Combinable) -> And:
        return And(self, value)

    def __or__(self, value: Combinable) -> Or:
        return Or(self, value)


class Combination(CombinableMixin):

    operator = ''

    def __init__(self, first: Combinable, second: Combinable) -> None:
        self.first = first
        self.second = second

    @property
    def querystring(self):
        return QueryString(
            '({} ' + self.operator + ' {})',
            self.first.querystring,
            self.second.querystring
        )

    def __str__(self):
        self.querystring.__str__()


class And(Combination):
    operator = 'AND'


class Or(Combination):
    operator = 'OR'


class Where(CombinableMixin):

    def __init__(
        self,
        column: 'Column',
        value: t.Any = None,
        values: Iterable = [],
        operator: t.Type[Operator] = Operator
    ) -> None:
        self.column = column
        self.value = value
        self.values = values
        self.operator = operator

    @property
    def values_querystring(self) -> QueryString:
        template = ', '.join(['{}' for i in self.values])
        return QueryString(
            template,
            *self.values
        )

    @property
    def querystring(self) -> QueryString:
        args: t.List[t.Any] = []
        if self.value:
            args.append(self.value)
        if self.values:
            args.append(self.values_querystring)

        # TODO Want something cleaner than this.
        template = self.operator.template.format(
            name=self.column.get_full_name(just_alias=True),
            value='{}',
            values='{}'
        )

        return QueryString(
            template,
            *args
        )

    def __str__(self):
        return self.querystring.__str__()



from __future__ import annotations

from piccolo.query.base import Query
from piccolo.querystring import QueryString


class Raw(Query):

    @property
    def querystring(self) -> QueryString:
        return self.base

    def __str__(self):
        return self.querystring.__str__()



from __future__ import annotations
from collections import OrderedDict
import typing as t

from piccolo.query.base import Query
from piccolo.columns import Column
from piccolo.query.mixins import (
    ColumnsMixin,
    CountMixin,
    DistinctMixin,
    LimitMixin,
    OrderByMixin,
    OutputMixin,
    WhereMixin,
)
from piccolo.querystring import QueryString

if t.TYPE_CHECKING:
    from table import Table  # noqa


class Select(
    Query,
    ColumnsMixin,
    CountMixin,
    DistinctMixin,
    LimitMixin,
    OrderByMixin,
    OutputMixin,
    WhereMixin,
):
    def get_joins(self, columns: t.List[Column]) -> t.List[str]:
        """
        A call chain is a sequence of foreign keys representing joins which
        need to be made to retrieve a column in another table.
        """
        joins: t.List[str] = []
        for column in columns:
            _joins: t.List[str] = []
            for index, key in enumerate(column.call_chain, 0):
                table_alias = "$".join(
                    [
                        f"{_key._table.Meta.tablename}${_key._name}"
                        for _key in column.call_chain[: index + 1]
                    ]
                )
                key.table_alias = table_alias

                if index > 0:
                    left_tablename = column.call_chain[index - 1].table_alias
                else:
                    left_tablename = key._table.Meta.tablename

                _joins.append(
                    f"JOIN {key.references.Meta.tablename} {table_alias} "
                    f"ON ({left_tablename}.{key._name} = {table_alias}.id)"
                )

            joins.extend(_joins)

        # Remove duplicates
        return list(OrderedDict.fromkeys(joins))

    def check_valid_call_chain(self, keys: t.List[Column]) -> bool:
        for column in keys:
            if column.call_chain:
                # Make sure the call_chain isn't too large to discourage
                # very inefficient queries.

                if len(column.call_chain) > 10:
                    raise Exception(
                        "Joining more than 10 tables isn't supported - "
                        "please restructure your query."
                    )
        return True

    @property
    def querystring(self) -> QueryString:
        # JOIN
        self.check_valid_call_chain(self.selected_columns)

        select_joins = self.get_joins(self.selected_columns)
        where_joins = self.get_joins(self.get_where_columns())
        order_by_joins = self.get_joins(self.get_order_by_columns())

        # Combine all joins, and remove duplicates
        joins: t.List[str] = list(
            OrderedDict.fromkeys(
                select_joins + where_joins + order_by_joins
            )
        )

        #######################################################################

        if len(self.selected_columns) == 0:
            self.selected_columns = self.table.Meta.columns

        column_names: t.List[str] = [
            c.get_full_name() for c in self.selected_columns
        ]
        columns_str = ", ".join(column_names)

        #######################################################################

        select = "SELECT DISTINCT" if self.distinct else "SELECT"
        query = f'{select} {columns_str} FROM {self.table.Meta.tablename}'

        for join in joins:
            query += f" {join}"

        #######################################################################

        args: t.List[t.Any] = []

        if self._where:
            query += " WHERE {}"
            args.append(self._where.querystring)

        if self._order_by:
            query += " {}"
            args.append(self._order_by.querystring)

        if self._limit:
            query += " {}"
            args.append(self._limit.querystring)

        querystring = QueryString(query, *args)

        if self._count:
            querystring = QueryString(
                "SELECT COUNT(*) FROM ({}) AS sub_query",
                querystring
            )

        return querystring

    def __str__(self) -> str:
        return self.querystring.__str__()



from .column_types import (  # noqa
    Varchar,
    Integer,
    Serial,
    PrimaryKey,
    Timestamp,
    Boolean,
    ForeignKey
)
from .base import Column  # noqa
from .combination import And, Or, Where  # noqa



from ..base import DBTestCase
from ..example_project.tables import Band


class TestDelete(DBTestCase):

    def test_delete(self):
        self.insert_rows()

        Band.delete.where(
            Band.name == 'CSharps'
        ).run_sync()

        response = Band.select.where(
            Band.name == 'CSharps'
        ).count().run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'count': 0}]
        )



from unittest import TestCase

from ..example_project.tables import Manager


class TestTableExists(TestCase):

    def setUp(self):
        Manager.create.run_sync()

    def test_table_exists(self):
        response = Manager.table_exists.run_sync()
        self.assertTrue(response is True)

    def tearDown(self):
        Manager.drop.run_sync()



"""
Used to generate and modify database tables.

How can I make tables easily diffable???

Work from a theoretical picture of how the database should look, or work from
the actual db???

Work from theoretical ...

AddColumn / DropColumn / ModifyColumn all need to be combinable.

Also, the attributes also need to be combinable.

Rather than them being directly combinable, they just manipulate a schema
object. Add and delete operations are processed first.

Just make the Schema's combinable.

"""
from __future__ import annotations
# TODO - not used much yet. Remove???

# Consists of a bunch of operators
# create_column
# create_table
# modify_column (rename, null / not null, max_chars, change type)
# drop_column
# We already have alter ...


class Schema():

    def __init__(self, column_changes, table_changes):
        self.column_changes = column_changes
        self.table_changes = table_changes

    def __add__(self, value: Schema) -> Schema:
        """
        We add schema objects together to get a picture of the schema at
        a point in time.
        """
        self.column_changes += value.column_changes
        self.table_changes += value.table_changes
        return self


class CreateTable():
    def __init__(self):
        pass


class ModifyTable():
    """
    Does things like renaming the table.
    """
    def __init__(self):
        pass


class CreateColumn():
    def __init__(self):
        pass


class ModifyColumn():
    def __init__(self):
        pass


class DropColumn():
    def __init__(self, table_name: str, column_name: str) -> None:
        self.table_name = table_name
        self.column_name = column_name

    def __str__(self):
        return f'DROP COLUMN {self.column_name} from {self.table_name};'



from .base import Query
from .methods import (
    Alter,
    Select,
    Objects,
    Insert,
    Delete,
    Create,
    Update,
    Raw,
    Drop,
    TableExists,
    Exists,
)


__all__ = [
    'Alter',
    'Query',
    'Select',
    'Objects',
    'Insert',
    'Delete',
    'Create',
    'Update',
    'Raw',
    'Drop',
    'TableExists',
    'Exists',
]



from __future__ import annotations
import itertools
import typing as t

import ujson as json

from piccolo.querystring import QueryString
from piccolo.utils.sync import run_sync

if t.TYPE_CHECKING:
    from table import Table  # noqa


class Query(object):

    def __init__(
        self,
        table: Table,
        base: QueryString = QueryString(''),
        *args,
        **kwargs
    ) -> None:
        self.base = base
        self.table = table
        super().__init__()

    @property
    def engine_type(self) -> str:
        return self.table.Meta.db.engine_type

    async def run(self, as_dict=True, in_pool=False):
        engine = getattr(self.table.Meta, 'db', None)
        if not engine:
            raise ValueError(
                f'Table {self.table.Meta.tablename} has no db defined in Meta'
            )

        results = await engine.run_querystring(
            self.querystring,
            in_pool=in_pool
        )

        if results:
            keys = results[0].keys()
            keys = [i.replace('$', '.') for i in keys]
            raw = [dict(zip(keys, i.values())) for i in results]
        else:
            raw = []

        if hasattr(self, 'run_callback'):
            self.run_callback(raw)

        # TODO - I have multiple ways of modifying the final output
        # response_handlers, and output ...
        # Might try and merge them.
        raw = self.response_handler(raw)

        output = getattr(self, '_output', None)

        if output:
            if output.as_objects:
                # When using .first() we get a single row, not a list
                # of rows.
                if type(raw) is list:
                    raw = [self.table(**columns) for columns in raw]
                else:
                    raw = self.table(**raw)
            elif type(raw) is list:
                if output.as_list:
                    if len(raw[0].keys()) != 1:
                        raise ValueError(
                            'Each row returned more than one value'
                        )
                    else:
                        raw = list(
                            itertools.chain(*[j.values() for j in raw])
                        )
                if output.as_json:
                    raw = json.dumps(raw)

        return raw

    def run_sync(self, *args, **kwargs):
        """
        A convenience method for running the coroutine synchronously.

        Might make it more sophisticated in the future, so not creating /
        tearing down connections, but instead running it in a separate
        process, and dispatching coroutines to it.
        """
        return run_sync(
            self.run(*args, **kwargs, in_pool=False)
        )

    def response_handler(self, response):
        """
        Subclasses can override this to modify the raw response returned by
        the database driver.
        """
        return response

    ###########################################################################

    @property
    def sqlite_querystring(self) -> QueryString:
        raise NotImplementedError

    @property
    def postgres_querystring(self) -> QueryString:
        raise NotImplementedError

    @property
    def default_querystring(self) -> QueryString:
        raise NotImplementedError

    @property
    def querystring(self) -> QueryString:
        """
        Calls the correct underlying method, depending on the current engine.
        """
        engine_type = self.engine_type
        if engine_type == 'postgres':
            try:
                return self.postgres_querystring
            except NotImplementedError:
                return self.default_querystring
        elif engine_type == 'sqlite':
            try:
                return self.sqlite_querystring
            except NotImplementedError:
                return self.default_querystring
        else:
            raise Exception(
                f'No querystring found for the {engine_type} engine.'
            )

    ###########################################################################

    def __str__(self) -> str:
        return self.querystring.__str__()



from ..base import DBTestCase
from ..example_project.tables import Band, Concert


class TestSelect(DBTestCase):

    def test_query_all_columns(self):
        self.insert_row()

        response = Band.select.run_sync()
        print(f'response = {response}')

        self.assertDictEqual(
            response[0],
            {
                'id': 1,
                'name': 'Pythonistas',
                'manager': 1,
                'popularity': 1000
            }
        )

    def test_query_some_columns(self):
        self.insert_row()

        response = Band.select.columns(Band.name).run_sync()
        print(f'response = {response}')

        self.assertDictEqual(
            response[0],
            {'name': 'Pythonistas'}
        )

    def test_where_like(self):
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).where(
            Band.name.like('Python%')
        ).run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'Pythonistas'}]
        )

    def test_where_not_like(self):
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).where(
            Band.name.not_like('Python%')
        ).order_by(
            Band.name
        ).run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'CSharps'}, {'name': 'Rustaceans'}]
        )

    def test_where_greater_than(self):
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).where(
            Band.popularity > 1000
        ).run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'Rustaceans'}]
        )

    def test_where_greater_equal_than(self):
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).where(
            Band.popularity >= 1000
        ).order_by(
            Band.name
        ).run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'Pythonistas'}, {'name': 'Rustaceans'}]
        )

    def test_where_less_than(self):
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).where(
            Band.popularity < 1000
        ).run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'CSharps'}]
        )

    def test_where_less_equal_than(self):
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).where(
            Band.popularity <= 1000
        ).run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'Pythonistas'}, {'name': 'CSharps'}]
        )

    def test_where_and(self):
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).where(
            (Band.popularity <= 1000) & (Band.name.like('Python%'))
        ).run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'Pythonistas'}]
        )

    def test_where_or(self):
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).where(
            (Band.name == 'Rustaceans') | (Band.name == 'CSharps')
        ).order_by(
            Band.name
        ).run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'CSharps'}, {'name': 'Rustaceans'}]
        )

    def test_multiple_where(self):
        """
        Test that chaining multiple where clauses works results in an AND.
        """
        self.insert_rows()

        query = Band.select.columns(
            Band.name
        ).where(
            Band.name == 'Rustaceans'
        ).where(
            Band.manager == 2
        )

        response = query.run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'Rustaceans'}]
        )
        self.assertTrue(
            'AND' in query.__str__()
        )

    def test_complex_where(self):
        """
        Test a complex where clause - combining AND, and OR.
        """
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).where(
            ((Band.popularity == 2000) & (Band.manager == 2)) |
            ((Band.popularity == 10) & (Band.manager == 3))
        ).order_by(
            Band.name
        ).run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'CSharps'}, {'name': 'Rustaceans'}]
        )

    def test_limit(self):
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).order_by(
            Band.name
        ).limit(
            1
        ).run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'CSharps'}]
        )

    def test_first(self):
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).order_by(
            Band.name
        ).first.run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            {'name': 'CSharps'}
        )

    def test_order_by_ascending(self):
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).order_by(
            Band.name
        ).limit(1).run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'CSharps'}]
        )

    def test_order_by_decending(self):
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).order_by(
            Band.name,
            ascending=False
        ).limit(1).run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'name': 'Rustaceans'}]
        )

    def test_count(self):
        self.insert_rows()

        response = Band.select.where(
            Band.name == 'Pythonistas'
        ).count().run_sync()

        print(f'response = {response}')

        self.assertEqual(
            response,
            [{'count': 1}]
        )

    def test_distinct(self):
        self.insert_rows()
        self.insert_rows()

        response = Band.select.columns(
            Band.name
        ).where(
            Band.name == 'Pythonistas'
        ).distinct().run_sync()

        self.assertTrue(response == [{'name': 'Pythonistas'}])

    def test_call_chain(self):
        """
        Make sure the call chain lengths are the correct size.
        """
        self.assertEqual(
            len(
                Concert.band_1.name.call_chain
            ),
            1
        )
        self.assertEqual(
            len(
                Concert.band_1.manager.name.call_chain
            ),
            2
        )



class Operator():
    template = ''


class Equal(Operator):
    template = '{name} = {value}'


class NotEqual(Operator):
    template = '{name} != {value}'


class In(Operator):
    template = '{name} IN ({values})'


class NotIn(Operator):
    template = '{name} NOT IN ({values})'


class Like(Operator):
    template = "{name} LIKE {value}"


class NotLike(Operator):
    template = "{name} NOT LIKE {value}"


class GreaterThan(Operator):
    # Add permitted types???
    template = "{name} > {value}"


class GreaterEqualThan(Operator):
    template = "{name} >= {value}"


class LessThan(Operator):
    template = "{name} < {value}"


class LessEqualThan(Operator):
    template = "{name} <= {value}"



import asyncio
from unittest import TestCase

from piccolo.extensions.user import User
from ..example_project.tables import DB


class _User(User):
    class Meta():
        tablename = 'a_user'
        db = DB


class TestCreateUserTable(TestCase):

    def test_create_user_table(self):
        """
        Make sure the table can be created.
        """
        exception = None
        try:
            _User.create.run_sync()
        except Exception as e:
            exception = e
        else:
            _User.drop.run_sync()

        if exception:
            raise exception

        self.assertFalse(exception)


class TestHashPassword(TestCase):

    def test_hash_password(self):
        pass


class TestLogin(TestCase):

    def setUp(self):
        _User.create.run_sync()

    def tearDown(self):
        _User.drop.run_sync()

    def test_login(self):
        username = "bob"
        password = "Bob123$$$"
        email = "bob@bob.com"

        save_query = _User(
            username=username,
            password=password,
            email=email
        ).save

        save_query.run_sync()

        authenticated = asyncio.run(
            _User.login(username, password)
        )
        self.assertTrue(authenticated is not None)

        authenticated = asyncio.run(
            _User.login(username, 'blablabla')
        )
        self.assertTrue(not authenticated)



import asyncio
import datetime
import importlib.util
import os
import sys
import typing as t
from types import ModuleType

import click

from piccolo.commands.playground import playground
from piccolo.engine import PostgresEngine
from piccolo.migrations.template import TEMPLATE
from piccolo.migrations.table import Migration


MIGRATIONS_FOLDER = os.path.join(os.getcwd(), 'migrations')
MIGRATION_MODULES: t.Dict[str, ModuleType] = {}


def _create_migrations_folder() -> bool:
    """
    Creates the folder that migrations live in. Returns True/False depending
    on whether it was created or not.
    """
    if os.path.exists(MIGRATIONS_FOLDER):
        return False
    else:
        os.mkdir(MIGRATIONS_FOLDER)
        for filename in ('__init__.py', 'config.py'):
            with open(os.path.join(MIGRATIONS_FOLDER, filename), 'w'):
                pass
        return True


def _create_new_migration() -> None:
    """
    Creates a new migration file on disk.
    """
    _id = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
    path = os.path.join(MIGRATIONS_FOLDER, f'{_id}.py')
    with open(path, 'w') as f:
        f.write(TEMPLATE.format(migration_id=_id))


###############################################################################

@click.command()
def new():
    """
    Creates a new file like migrations/0001_add_user_table.py
    """
    print('Creating new migration ...')
    _create_migrations_folder()
    _create_new_migration()


###############################################################################

def _create_migration_table() -> bool:
    """
    Creates the migration table in the database. Returns True/False depending
    on whether it was created or not.
    """
    if not Migration.table_exists.run_sync():
        Migration.create.run_sync()
        return True
    return False


def _get_migrations_which_ran() -> t.List[str]:
    """
    Returns the names of migrations which have already run, by inspecing the
    database.
    """
    return [
        i['name'] for i in Migration.select.columns(Migration.name).run_sync()
    ]


def _get_migration_modules() -> None:
    """
    """
    folder_contents = os.listdir(MIGRATIONS_FOLDER)
    excluded = ('__init__.py', 'config.py', '__pycache__')
    migration_names = [
        i.split('.py')[0] for i in folder_contents if i not in excluded
    ]
    modules = [importlib.import_module(name) for name in migration_names]
    global MIGRATION_MODULES
    for m in modules:
        _id = getattr(m, 'ID', None)
        if _id:
            MIGRATION_MODULES[_id] = m


def _get_migration_ids() -> t.List[str]:
    return list(MIGRATION_MODULES.keys())


def _get_config() -> dict:
    """
    A config file is required for the database credentials.
    """
    sys.path.insert(0, MIGRATIONS_FOLDER)

    config_file = os.path.join(MIGRATIONS_FOLDER, 'config.py')
    if not os.path.exists(config_file):
        raise Exception(f"Can't find config.py in {MIGRATIONS_FOLDER}")

    config = importlib.import_module('config')

    db = getattr(config, 'DB', None)
    if not db:
        raise Exception('config.py is missing a DB dictionary.')
    return db


@click.command()
def forwards():
    """
    Runs any migrations which haven't been run yet, or up to a specific
    migration.
    """
    print('Running migrations ...')
    sys.path.insert(0, os.getcwd())

    Migration.Meta.db = PostgresEngine(_get_config())

    _create_migration_table()

    already_ran = _get_migrations_which_ran()
    print(f'Already ran = {already_ran}')

    # TODO - stop using globals ...
    _get_migration_modules()
    ids = _get_migration_ids()
    print(f'Migration ids = {ids}')

    for _id in (set(ids) - set(already_ran)):
        asyncio.run(
            MIGRATION_MODULES[_id].forwards()
        )

        print(f'Ran {_id}')
        Migration.insert().add(
            Migration(name=_id)
        ).run_sync()


###############################################################################

@click.command()
@click.argument('migration_name')
def backwards(migration_name: str):
    """
    Undo migrations up to a specific migrations.

    - make sure the migration name is valid
    - work out which to undo, and in which order
    - ask for confirmation
    - apply the undo operations one by one
    """
    # Get the list from disk ...
    sys.path.insert(0, os.getcwd())
    _get_config()  # Just required for path manipulation - needs changing
    _get_migration_modules()

    Migration.Meta.db = PostgresEngine(_get_config())

    _create_migration_table()

    migration_ids = _get_migration_ids()

    if migration_name not in migration_ids:
        print(f'Unrecognized migration name - must be one of {migration_ids}')

    _continue = input('About to undo the migrations - press y to continue.')
    if _continue == 'y':
        print('Undoing migrations')

        _sorted = sorted(list(MIGRATION_MODULES.keys()))
        _sorted = _sorted[_sorted.index(migration_name):]
        _sorted.reverse()

        already_ran = _get_migrations_which_ran()

        for s in _sorted:
            if s not in already_ran:
                print(f"Migration {s} hasn't run yet, can't undo!")
                sys.exit(1)

            print(migration_name)
            asyncio.run(
                MIGRATION_MODULES[s].backwards()  # type: ignore
            )

            Migration.delete.where(
                Migration.name == s
            ).run_sync()
    else:
        print('Not proceeding.')


###############################################################################

@click.group('migration')
def cli():
    pass


cli.add_command(new)
cli.add_command(forwards)
cli.add_command(backwards)
cli.add_command(playground)


def main():
    cli()


###############################################################################


if __name__ == '__main__':
    main()



import sqlite3
import typing as t

import aiosqlite

from piccolo.engine.base import Engine
from piccolo.querystring import QueryString


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


class SQLiteEngine(Engine):

    engine_type = 'sqlite'

    def __init__(self, path: str = 'piccolo.sqlite') -> None:
        self.path = path

    async def run_in_pool(
        self,
        query: str,
        args: t.List[t.Any] = [],
        query_type: str = 'generic'
    ):
        raise NotImplementedError

    async def run(
        self,
        query: str,
        args: t.List[t.Any] = [],
        query_type: str = 'generic'
    ):
        async with aiosqlite.connect(self.path) as connection:
            connection.row_factory = dict_factory
            async with connection.execute(query, args) as cursor:
                cursor.row_factory = dict_factory
                await connection.commit()
                response = await cursor.fetchall()
                print(query)
                print(args)
                print(response)

                print(query_type)
                if query_type == 'insert':
                    return [{'id': cursor.lastrowid}]
                else:
                    return response

    async def run_querystring(
        self,
        querystring: QueryString,
        in_pool: bool = False
    ):
        if in_pool:
            return await self.run_in_pool(
                *querystring.compile_string(engine_type=self.engine_type),
                query_type=querystring.query_type
            )
        else:
            return await self.run(
                *querystring.compile_string(engine_type=self.engine_type),
                query_type=querystring.query_type
            )

    def transaction(self):
        raise NotImplementedError



from unittest import TestCase

from ..example_project.tables import Band, Manager, Concert, Venue


TABLES = [Manager, Band, Venue, Concert]


class TestCreateJoin():

    def test_create_join(self):
        for table in TABLES:
            table.create.run_sync()

        for table in reversed(TABLES):
            table.drop.run_sync()


class TestJoin(TestCase):
    """
    Test instantiating Table instances
    """

    tables = [Manager, Band, Venue, Concert]

    def setUp(self):
        for table in self.tables:
            table.create.run_sync()

    def tearDown(self):
        for table in reversed(self.tables):
            table.drop.run_sync()

    def test_join(self):
        manager_1 = Manager(name="Guido")
        manager_1.save.run_sync()

        band_1 = Band(name="Pythonistas", manager=manager_1.id)
        band_1.save.run_sync()

        manager_2 = Manager(name="Graydon")
        manager_2.save.run_sync()

        band_2 = Band(name="Rustaceans", manager=manager_2.id)
        band_2.save.run_sync()

        venue = Venue(name="Grand Central")
        venue.save.run_sync()

        # TODO - make sure you can also do:
        # band_1=Pythonistas
        save_query = Concert(
            band_1=band_1.id,
            band_2=band_2.id,
            venue=venue.id
        ).save
        save_query.run_sync()

        select_query = Concert.select.columns(
            Concert.band_1.name,
            Concert.band_2.name,
            Concert.venue.name,
            Concert.band_1.manager
        )
        response = select_query.run_sync()
        print(response)

    # def _test_ref(self):
    #     """
    #     Concert.select().count().where(
    #         Concert.ref('band1.name') == 'Pythonistas'
    #     )
    #     """
    #     pass



"""
A User model which can be subclassed in projects.
"""
import hashlib
import secrets
import typing as t

from piccolo.table import Table
from piccolo.columns import Varchar, Boolean


# Might change to BaseUser ... making it clearer that it needs to be
# subclassed ...
class User(Table):
    """
    The password needs to be hashed.
    """
    username = Varchar(length=100)
    password = Varchar(length=255)
    email = Varchar(length=255)
    active = Boolean(default=False)

    class Meta():
        tablename = 'a_user'

    def __init__(self, **kwargs):
        """
        Generating passwords upfront is expensive, so might need reworking.
        """
        password = kwargs.get('password', None)
        if password:
            kwargs['password'] = self.__class__.hash_password(password)
        super().__init__(**kwargs)

    @classmethod
    def get_salt(cls):
        return secrets.token_hex(16)

    @classmethod
    def hash_password(
        cls,
        password: str,
        salt: str = '',
        iterations: int = 10000
    ) -> str:
        """
        Hashes the password, ready for storage, and for comparing during
        login.
        """
        if salt == '':
            salt = cls.get_salt()
        hashed = hashlib.pbkdf2_hmac(
            'sha256',
            bytes(password, encoding="utf-8"),
            bytes(salt, encoding="utf-8"),
            iterations
        ).hex()
        return f'pbkdf2_sha256${iterations}${salt}${hashed}'

    @classmethod
    def split_stored_password(self, password: str) -> t.List[str]:
        elements = password.split('$')
        if len(elements) != 4:
            raise ValueError('Unable to split hashed password')
        return elements

    @classmethod
    async def login(cls, username: str, password: str) -> t.Optional[int]:
        """
        Returns the user_id if a match is found.
        """
        try:
            response = await cls.select.columns(cls.id, cls.password).where(
                (cls.username == username)
            ).first.run()
        except ValueError:
            # No match found
            return None

        stored_password = response['password']

        algorithm, iterations, salt, hashed = cls.split_stored_password(
            stored_password
        )

        if cls.hash_password(
            password,
            salt,
            int(iterations)
        ) == stored_password:
            return response['id']
        else:
            return None


# Things to consider ...
# just have some class methods around it ...
# or some methods ...
# One problem is ... need to define a secret key to salt the hash
# When people subclass it ... can add the hash???
# Need to have the ability in a table to transform values during __init__ and
# __set__.

# I don't think you have one salt ... I think it's random each time ...
# and you use it to hash the password, and then store something like this:
# encryption_method$iterations$salt$hashed_and_salted_password
# If you know the salt ... couldn't you just regenerate rainbow tables???



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import find_packages, setup

from piccolo import __VERSION__ as VERSION


directory = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(directory, 'requirements.txt')) as f:
    contents = f.read()
    REQUIREMENTS = [i.strip() for i in contents.strip().split('\n')]


with open(os.path.join(directory, 'README.md')) as f:
    LONG_DESCRIPTION = f.read()


setup(
    name='piccolo',
    version=VERSION,
    description='A fast, user friendly ORM which supports asyncio.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Daniel Townsend',
    author_email='dan@dantownsend.co.uk',
    python_requires='>=3.7.0',
    url='https://github.com/piccolo-orm/piccolo',
    packages=find_packages(exclude=('tests',)),
    install_requires=REQUIREMENTS,
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    entry_points={
        'console_scripts': [
            'piccolo = piccolo.main:main',
        ],
    },
)



from ..base import DBTestCase
from ..example_project.tables import Band


class TestObjects(DBTestCase):

    def test_get_all(self):
        self.insert_row()

        response = Band.objects.run_sync()

        self.assertTrue(len(response) == 1)

        instance = response[0]

        self.assertTrue(isinstance(instance, Band))
        self.assertTrue(instance.name == 'Pythonistas')

        # Now try changing the value and saving it.
        instance.name = 'Rustaceans'
        save_query = instance.save
        save_query.run_sync()

        self.assertTrue(
            Band.select.columns(
                Band.name
            ).output(
                as_list=True
            ).run_sync()[0] == 'Rustaceans'
        )



from __future__ import annotations
import datetime
from dataclasses import dataclass
from string import Formatter
import typing as t


@dataclass
class Unquoted():
    """
    Used when we want the value to be unquoted because it's a Postgres
    keyword - for example DEFAULT.
    """
    value: str


@dataclass
class Fragment():
    prefix: str
    index: int = 0
    no_arg: bool = False


class QueryString():

    def __init__(self, template: str, *args: t.Any, query_type: str = 'generic') -> None:
        """
        Example template: "WHERE {} = {}"

        The query type is sometimes used by the engine to modify how the query
        is run.
        """
        self.template = template
        self.args = args
        self.query_type = query_type

    def __str__(self):
        """
        The SQL returned by the __str__ method isn't used directly in queries
        - it's just a usability feature.
        """
        _, bundled, combined_args = self.bundle(
            start_index=1,
            bundled=[],
            combined_args=[]
        )
        template = ''.join([
            fragment.prefix + ('' if fragment.no_arg else '{}') for fragment in bundled
        ])

        # Do some basic type conversion here.
        converted_args = []
        for arg in combined_args:
            _type = type(arg)
            if _type == str:
                converted_args.append(f"'{arg}'")
            elif _type == datetime:
                dt_string = arg.isoformat().replace('T', ' ')
                converted_args.append(f"'{dt_string}'")
            elif arg == None:
                converted_args.append('null')
            else:
                converted_args.append(arg)

        return template.format(*converted_args)

    def bundle(self, start_index=1, bundled=[], combined_args=[]):
        # Split up the string, separating by {}.
        fragments = [
            Fragment(prefix=i[0]) for i in Formatter().parse(self.template)
        ]

        for index, fragment in enumerate(fragments):
            try:
                value = self.args[index]
            except IndexError:
                # trailing element
                fragment.no_arg = True
                bundled.append(fragment)
            else:
                if type(value) == self.__class__:
                    fragment.no_arg = True
                    bundled.append(fragment)

                    start_index, _, _ = value.bundle(
                        start_index=start_index,
                        bundled=bundled,
                        combined_args=combined_args,
                    )
                else:
                    fragment.index = start_index
                    bundled.append(fragment)
                    start_index += 1
                    combined_args.append(value)

        return (start_index, bundled, combined_args)

    def compile_string(
        self,
        engine_type: str = 'postgres'
    ) -> t.Tuple[str, t.List[t.Any]]:
        """
        Compiles the template ready for Postgres - keeping the arguments
        separate from the template.
        """
        _, bundled, combined_args = self.bundle(
            start_index=1,
            bundled=[],
            combined_args=[]
        )
        if engine_type == 'postgres':
            string = ''.join([
                fragment.prefix + ('' if fragment.no_arg else f'${fragment.index}') for fragment in bundled
            ])
        elif engine_type == 'sqlite':
            string = ''.join([
                fragment.prefix + ('' if fragment.no_arg else '?') for fragment in bundled
            ])
        else:
            raise Exception('Engine type not recognised')

        return (string, combined_args)



from unittest import TestCase

from tests.example_project.tables import Manager, Band


TABLES = [Manager, Band]


class TestGetRelated(TestCase):

    def setUp(self):
        for table in TABLES:
            table.create.run_sync()

    def tearDown(self):
        for table in reversed(TABLES):
            table.drop.run_sync()

    def test_get_related(self):
        """
        Make sure you can get a related object from another object instance.
        """
        manager = Manager(
            name="Guido"
        )
        manager.save.run_sync()

        band = Band(
            name='Pythonistas',
            manager=manager.id,
            popularity=100
        )
        band.save.run_sync()

        _manager = band.get_related('manager').run_sync()

        self.assertTrue(
            _manager.name == 'Guido'
        )



from __future__ import annotations

from piccolo.query.base import Query
from piccolo.querystring import QueryString


class Drop(Query):

    @property
    def querystring(self) -> QueryString:
        return QueryString(
            f'DROP TABLE "{self.table.Meta.tablename}"'
        )

    def __str__(self) -> str:
        return self.querystring.__str__()



"""
Populates a database with an example schema and data, and launches a shell
for interacting with the data using Piccolo.
"""
import sys

import click

from piccolo import table
from piccolo import columns
from piccolo.engine import SQLiteEngine, PostgresEngine


# DB = PostgresEngine({})
DB = SQLiteEngine()


class Manager(table.Table):
    name = columns.Varchar(length=50)

    class Meta():
        db = DB


class Band(table.Table):
    name = columns.Varchar(length=50)
    manager = columns.ForeignKey(references=Manager)
    popularity = columns.Integer()

    class Meta():
        db = DB


class Venue(table.Table):
    name = columns.Varchar(length=100)
    capacity = columns.Integer()

    class Meta():
        db = DB


class Concert(table.Table):
    band_1 = columns.ForeignKey(Band)
    band_2 = columns.ForeignKey(Band)
    venue = columns.ForeignKey(Venue)

    class Meta():
        db = DB


TABLES = (Manager, Band, Venue, Concert)


def populate():
    """
    Drop then recreate the tables, and populate with data.
    """
    for _table in reversed(TABLES):
        try:
            _table.drop.run_sync()
        except Exception as e:
            print(e)

    for _table in TABLES:
        try:
            _table.create.run_sync()
        except Exception as e:
            print(e)

    guido = Manager(name="Guido")
    guido.save.run_sync()

    pythonistas = Band(
        name="Pythonistas",
        manager=guido.id,
        popularity=1000
    )
    pythonistas.save.run_sync()

    graydon = Manager(name="Graydon")
    graydon.save.run_sync()

    rustaceans = Band(
        name="Rustaceans",
        manager=graydon.id,
        popularity=500
    )
    rustaceans.save.run_sync()

    venue = Venue(
        name="Amazing Venue",
        capacity=5000
    )
    venue.save.run_sync()

    concert = Concert(
        band_1=pythonistas.id,
        band_2=rustaceans.id,
        venue=venue.id
    )
    concert.save.run_sync()


@click.command(name="playground")
@click.option('--user', default='piccolo', help='Postgres user')
@click.option('--password', default='piccolo', help='Postgres password')
@click.option('--database', default='piccolo_playground',
              help='Postgres database')
@click.option('--host', default='localhost', help='Postgres host')
@click.option('--port', default=5432, help='Postgres port')
def playground(user, password, database, host, port):
    """
    Creates a test database to play with.
    """
    try:
        import IPython
    except ImportError:
        print(
            "Install iPython using `pip install ipython==7.1.1` to use this "
            "feature."
        )
        sys.exit(1)

    global DB

    DB.config = {
        'host': host,
        'database': database,
        'user': user,
        'password': password,
        'port': port
    }

    print('Tables:\n')

    for _table in TABLES:
        print(_table)
        print('\n')

    populate()
    IPython.embed()



from unittest import TestCase

from piccolo.querystring import QueryString

from ..example_project.tables import Band


# TODO - add more extensive tests (increased nesting and argument count).
class TestQuerystring(TestCase):

    qs = QueryString(
        'SELECT id FROM band {}',
        QueryString(
            "WHERE name = {}",
            'Pythonistas'
        )
    )

    def test_compile_string(self):
        compiled_string, args = self.qs.compile_string()

        self.assertEqual(
            compiled_string,
            "SELECT id FROM band WHERE name = $1"
        )

        self.assertEqual(
            args,
            ['Pythonistas']
        )

    def test_string(self):
        string = self.qs.__str__()
        self.assertEqual(
            string,
            "SELECT id FROM band WHERE name = 'Pythonistas'"
        )

    def test_querystring_with_no_args(self):
        qs = QueryString(
            'SELECT name FROM band',
        )
        self.assertEqual(
            qs.compile_string(),
            ('SELECT name FROM band', [])
        )

# class TestExecuteQuerystring(TestCase):

#     def setUp():
#         Band.create.run_sync()

#     def tearDown():
#         Band.drop.run_sync()

#     def test_raw_query(self):
#         Band.Meta.db.run()
